# Econ 70310: Numerical Methods in Economics
# Fall 2017
# Lecture 1.1: Basics
# Example Code
# Zach Stangebye, Aug 2017

##
# Make sure to save all files as ".jl" files when working in the Juno IDE. It
# enables Juno to identify pre-built functions and keywords, making code more readable

# Comments in Julia are done using a "#" symbol. Anything following that symbol is not read

#= multiline comment
NONE OF THIS STUFF IS EXECUTED
=#
but_this_stuff_would_be = 1


# Declaring variables in Juno works like a dynamic language. Just state the variable
# and the value you want it to take

x = 1 # Numbers are automatically taken to be integers
y = 2.0 # Add a decimal to make it a float type

# Unlike Matlab don't need to add ';' to suppress output to console
# Instead, show things in the console with the println() function
println(x)
println(y)

# You can also see them in the Workspace





# Variables can also store characters and string names
# characters use single-quotation marks
mychar = 'h'

# strings use either 2 or 3 quotation marks
mystring = "hello there "
plus_yourstring = "world"

# Concatenation of strings can be done with the string function
ourstring = string(mystring,plus_yourstring)
println(ourstring)


# Juno automatically detects the most likely type
println(string("x is of type ",typeof(x)))
println(string("y is of type ",typeof(y)))
println(string("ourstring is of type ",typeof(ourstring)))

# Floats and Ints are automatically stored on a 64-bit for a 64-bit system

# Operations using two different types result in finer type
z = x + y
println(string("z is of type ",typeof(z)))

# If the operation does not make sense, though, Juno will give you an error
bad_move = z + ourstring

# Operations can span over more than one line if each line ends in an operator
# or if entire expression is closed in parentheses
z_other = x +
  y

z_other = (x
  + y)

# One cool feature of Julia is it intuitively distinguishes numbers from Variables
# in a polynomial expression
z_result1 = 2*x^2 + 5*x -2 # The standard numerical (Matlab) way
z_result2 = 2x^2 + 5x - 2 # The more intuitive mathematical way




# Can combine x and y using brackets: Becomes either a vector or a matrix depending
# on how you define it
vec1  = [x,y] # vectors can be defined with or without commas
mat1 = [x y; y x] # matrices use semicolons to start new row; no commas between columns

println(vec1)
println(mat1)

# Matrices
grab_bag_vec = [x, y, ourstring]
grab_bag_mat = [x y ourstring; y ourstring x]

# Access array elements with brackets
vec1_entry1 = vec1[1]
mat1_entry12 = mat1[1,2]

# Use a colon to access multiple elements
mat1_col2 = mat1[1:2,2]
mat1_row1 = mat1[1,1:2]

# Be careful with assigning arrays: Julia's default is to make them
# pointers.

A = rand(2,2)
B = A # B acquires the memory location of A, not just its values
# You can check if two variables share the same memory with ===
println(A === B)
A[1,1] = -.2
println(A === B)
# You'll now find the B has been altered as well. This problem only happens
# with assignment. Altering A mathematically doesn't change B. It's only
# assignment
A += 1
# B is unchanged by this. Alterations made to A following this
# operation also do not affect B since the memory is separate

# To avoid this problem, you can use copy()
A = rand(2,2)
B = copy(A)
println(A === B)
A[1,1] = -.2
# Now B is unaffected

# When dealing with 2D arrays, note that (like Matlab), * assumes matrix multiplication
# You can use .* to make it do element by elements
A = ones(2,2)
B = ones(2,2)
C_mat_mult = A*B # works as long as (MxN)*(NxP) -> MxP array
C_elmt_mult = A.*B # works as long as (MxN)*(MxN) -> MxN array

A[1,1] = .2
A[2,2] = -.3
# Julia also supports the Matlab syntax for matrix inversion i.e.
# A^{-1}B = A\B    and   A*B^{-1} = A/B
C_div1 = A\B
C_div2 = B/A
# Notice that these are transposes of each other

# Example functions: Can define in-line or in a block
squarethis(x) = x^2

function squarethat(x)
  return x^2
end

# Can pass in argument types to make it go a little faster, but it can cause
# errors if you're not careful
function squarethat_focused(x::Float64)
  return x^2
end

z1 = squarethat(2)
z2 = squarethat(2.0)

z3 = squarethat_focused(2.0)
bad_move = squarethat_focused(2) # can't feed Int64 arguments

# Subroutines in Julia
y = [1.0 1.0]
function squarethis_subroutine!(y, x)
  y[1] = x^2
  y[2] = 18
  return nothing
end




# Boolean types
# Define directly
bool1 = true
if bool1
  println("Yep, it is true")
else
  println("No, you are wrong. It is false.")
end

# Define conditionally
bool2 = 1 == 2
if bool2
  println("Yep, it is true")
else
  println("No, you are wrong. It is false.")
end

bool3 = (1 == 1) & (1 == 3) # and condition
bool4 = (1 == 1) | (1 == 3) # else condition

# use '~' before a boolean to denote its opposite
bool5 = true
println(~bool5)

myint = 1
if myint == 1
  println("My number is 1")
elseif myint == 2
  println("My number is 2")
else
  println("You don't know my number!")
end



# Loops
N = 10
this_array = zeros(1,N)
for i=1:N
  this_array[i] = i
end

# Can also nest loops. Note you can use 'i=1:N' or 'j in 1:N'
this_mat = zeros(N,N)
for i=1:N
  for j in 1:N
    this_mat[i,j] = i+j
  end
end

# Can iterate over more obscure arrays as well
for i in grab_bag_vec
  println(i)
end

# Can choose to count by alternative integers
for i=1:2:N-1
  this_array[i] = 100.0
end
# Or even backward
for i=N:-2:2
  this_array[i] = -100.0
end


# while loops act similarly but use conditionals
counter = 0
while counter <= 10
  println(counter)
  counter += 1
end



# Interval bisection to find the equilibrium price: Example
# another cool feature of Julia! You can use unicode characters (greek letters, chinese
# characters, etc.) as variable names! Just hit 'tab' after writing the latex equivalent
# e.g. \alpha + tab = α

tic()

# preferences
β_11 = 2.2
β_12 = 4.5
β_21 = .34
β_22 = .88

# endowments
e_11 = 4.6
e_12 = .23
e_21 = .9
e_22 = .02

# demand functions
c_12(p) = β_12/(β_11 + β_12)*(p*e_11 + (1-p)*e_12)/(1-p)
c_22(p) = β_22/(β_21 + β_22)*(p*e_21 + (1-p)*e_22)/(1-p)

# market clearing expression
excess_demand(p) = c_12(p) + c_22(p) - ( e_12 + e_22)

# Now we're ready to do interval bisection
pL = 0.0
pH = 1.0

ttol = 1e-6 # desired accuracy
alg_done = false

while ~alg_done
  new_p = (pL + pH)/2.0
  current_ed = excess_demand(new_p)

  if current_ed > 0.0
    pH = new_p
  else
    pL = new_p
  end

  if abs(current_ed) < ttol
    alg_done = true
  end
end

p_star = (pL + pH)/2.0
err_lv = excess_demand(p_star)

# What is p_star? Why is it the way it is?

toc()
ttime = toq() # Tells us how much time has passed since the tic() command

# toc() prints the time (in seconds); toq() stores it as a variable
# Julia almost always faster subsequent time because of JIT compilation
